# include <signal.h>
# include <errno.h>
# include <string.h>
# include <unistd.h>
# include <stdlib.h>
# include <termios.h>
# include <stdio.h>
# include <stropts.h>
# include <poll.h>
# include <fcntl.h>
# include <sys/wait.h>
# include <sys/types.h>

# define PTYM_NAME "/dev/ptmx"

extern int TOGGLE_KEY;
extern int cGetKey(unsigned char *arg);

struct termios org_termios;
struct winsize win_size;
pid_t pid;
int notitle = 0;
char *ptys_name;
extern char *ptsname (int);

void cleanup ()
{
    int i;
    extern int errno;

    i = errno;
    kill (pid, SIGKILL);
    errno = i;
}

void sigchld ()
{
    int status;

    waitpid (pid, &status, 0);
    if (WEXITSTATUS(status))
#ifdef DEBUG
      fprintf(stderr,"child exited with status %d\n", WEXITSTATUS(status)); 
#else
	;
#endif

}

void error_exit(char *s, int exitcode) {
  perror(s);
  exit(exitcode);
}

int stty_raw(int fd) /* set tty to RAW mode, returns -1 on error */
{
  struct termios temp;
  /* The "original" tty state is supposed to be in org_termios */
  (void) memcpy (&temp, &org_termios, sizeof (org_termios));

  temp.c_lflag &= ~(ICANON | ISIG | ECHO | IEXTEN);
  temp.c_oflag &= ~OPOST;
  temp.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  temp.c_cflag &= ~(CSIZE | PARENB);
  temp.c_cflag |= CS8;

  temp.c_cc[VMIN] = 1;
  temp.c_cc[VTIME] = 0;

  if (tcsetattr(fd, TCSAFLUSH, &temp) < 0)
    return -1;
  
  return 0;
}


void tty_reset() {
  /* resets tty to the original state, saved in org_termios */
  tcsetattr(0, TCSAFLUSH, &org_termios);
}

void main (int argc, char **argv) {
    char *stemp;
    int fdm, fds; /* File descriptors for master & slave */
    int inpflag =0, outpflag=0;
    int i, error=0;
    unsigned char c[1024];
    struct pollfd polled [2];

    /* Command line options processing */
    int ch;
    while ((ch = getopt(argc, argv, "nt:")) != EOF) {
	switch (ch) {
		case 'n':
			notitle = 1;
			break;
		case 't':
			TOGGLE_KEY = optarg[0];
			break;
		case '?':
			error++;
	}
	if (error) {
		fprintf(stderr, "usage: gkbd [-n] [-t TOGGLE_KEY] [command]\n");
		exit(2);			
	}
    }

    /* Save current terminal settings in org_tios */
   	if (tcgetattr (0, &org_termios) < 0) 
      	        error_exit("tcgetattr failed",1);

    	if (ioctl(0, TIOCGWINSZ, (char *) &win_size) < 0)
      	        error_exit ("failed to get window size...", 1);

    if ((fdm = open (PTYM_NAME, O_RDWR)) < 0) 
      error_exit ("cannot open master pty device",1);

    if (grantpt (fdm) <0)
      error_exit ("grantpt failed", 1);

    if (unlockpt (fdm) < 0) 
      error_exit ("unlockpt failed",1);

    if (!(stemp = ptsname (fdm))) {
      error_exit ("ptsname failed",1);
    }
    ptys_name = strdup(stemp);
    
    if ((pid = fork ()) < 0) 
      error_exit ("fork failed", 1);
    
    if (pid) {        /* parent */
      signal (SIGCHLD, sigchld);

     		if (stty_raw(0) == -1) 
       			error_exit("could not set tty to raw mode", 1); 
     		atexit(tty_reset);

      	while (1) {
	  int retpoll;
	  if (inpflag && outpflag) {
		cleanup();
		exit(0);
	  }
          polled[0].fd = 0;
          polled[0].events = (POLLRDNORM & POLLHUP);
          polled[1].fd = fdm;
          polled[1].events = (POLLRDNORM & POLLHUP);
          if ((retpoll = poll (&polled[0], 2, 9999999)) == -1) {
	    if ((errno != EAGAIN) && (errno != EINTR)) {
              cleanup ();
              perror ("poll failure");
              exit (3);
	    } 
	    else continue;
          }
          if (polled[0].revents) {
            if ((i = cGetKey (&c[0])) <= 0) 
           	inpflag = 1; 
            else {
              write(fdm, c, 1);
            }
          }
          if (polled[1].revents) {
            if ((i = read (fdm, c, 1024)) <= 0) 
		outpflag = 1;
            else {
                write (1, c, i);
            }
          }
      }
     }          /* end parent */
    else {        /* child */
      if (setsid () < 0) 
          error_exit ("setsid failed...",1);

      close(fdm);
      close(2);
      close(1);
      close(0);

        if ((fds = open (ptys_name, O_RDWR)) < 0) 
        error_exit ("cannot open slave pty device", 1);

        if (ioctl (fds, I_PUSH, "ptem") < 0) 
        error_exit ("cannot PUSH ptem", 1);

        if (ioctl (fds, I_PUSH, "ldterm") < 0) 
        error_exit ("cannot PUSH ldterm", 1);

        if (ioctl (fds, I_PUSH, "ttcompat") < 0) 
        error_exit ("cannot PUSH ttcompat", 1);

	if (tcsetattr (fds, TCSAFLUSH, &org_termios) < 0)
		error_exit("tcsetattr failure on slave", 1);

	if (ioctl (fds, TIOCSWINSZ, &win_size) < 0)
  		error_exit("tiocswinsz failure on slave", 1);

	for (i = 0; i <= 2; i++) {  /* foreach of stdin, stdout, stderr */
	  if (dup2(fds,i) != i)   /* reopen it as a dup of slave */
	    exit (3);
	}
      
	if (optind == argc)
		execl(getenv("SHELL"), getenv("SHELL"), (char *) 0);
        else execvp (argv[optind], &argv[optind]);
  }          /* end child */
}

